#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Script to filter mass mail sendings, and handle quota limitation by periods
#
# Why ?
# Because postfwd is not the correct software to do this (handle period based
# on the first mail sending date)
#
# Guillaume Penaud (gpenaud@easter-eggs.com)
# 05/2015
# 
# exit codes
#   - 0: successful
#   - 1: nothing done
#   - 2: ldap server error
#

"""
Handle quota limitations on postfix
"""
import argparse
import logging
import os
import policy
import re
import signal
import socket
import sys
import yaml

from threading import Thread
from logging import handlers

from policy.api_manager import ApiManager
from policy.data_manager import DataManager
from policy.ldap_manager import LdapManager

# Server
# -----------------------------------------------------------------------------


class Server(object):

    config_filepath = "/etc/policy_manager/configuration.yaml"

    logger = None
    syslog_handler = None
    stderr_handler = None

    config = {}
    args   = {}

    def __init__(self):
        # mixed variables declaration
        self.threads = []
        self.stopped = False
        self.attrs = {}

# Load configuration object
# -----------------------------------------------------------------------------
    def load_config(self):
        config_filepath = self.args.get("configuration_file")

        if config_filepath is None:
            config_filepath = self.config_filepath

        # handle configuration
        try:
            stream = open(config_filepath, "r")
        except IOError:
            sys.stderr.write('[ERROR] Impossible to load configuration file\n')
            sys.exit(1)

        # set-up server configuration
        self.config = yaml.load(stream)

        # load logger here to be able to change configuration upon it
        self.load_logger()

        self.hostname = self.config.get("global", {}).get("policy_host", "127.0.0.1")
        self.port = self.config.get("global", {}).get("policy_port", 9001)

        # handle periods
        self.periods = {}

        available_periods = ["hourly", "daily", "weekly", "monthly", "yearly"]
        periods = self.config.get("period")

        if not all(period in available_periods for period in periods.keys()):
            self.logger.error('One of the period set is unauthorized')
            sys.exit(2)

        sorted_periods = periods.items()
        sorted_periods.sort(key=lambda v: v[1]["nb_seconds"])
        sorted_periods.reverse()

        for label, data in sorted_periods:
            self.periods.setdefault(label, data)


# Set-up custom logger
# -----------------------------------------------------------------------------

    def load_logger(self):
        self.logger = logging.getLogger('policy_manager')

        if self.syslog_handler is not None:
            self.logger.removeHandler(self.syslog_handler)

        if self.stderr_handler is not None:    
            self.logger.removeHandler(self.stderr_handler)
        
        log_level = self.config.get("global", {}).get("log_level", 10)
        self.logger.setLevel(log_level)

        mail_facility = logging.handlers.SysLogHandler.LOG_MAIL

        self.syslog_handler = logging.handlers.SysLogHandler(facility=mail_facility, address='/dev/log')
        self.stderr_handler = logging.StreamHandler(sys.stderr)

        # to have a nice log format
        formatter = logging.Formatter('%(name)s: %(levelname)s: %(message)s')

        self.syslog_handler.setFormatter(formatter)
        self.stderr_handler.setFormatter(formatter)

        self.logger.addHandler(self.syslog_handler)
        self.logger.addHandler(self.stderr_handler)


# Handle client connections
# -----------------------------------------------------------------------------

    def handle_connection(self, connection, address):
        self.logger.debug("Handle socket connection")
        try:
            while True:
                data = connection.recv(8192)
                # data = connection.recv(18192)

                if data == "":
                    self.logger.debug("Socket has been closed remotely")
                    break

                attrs = self.parse_request(connection, data)

                if "request" not in attrs:
                    self.logger.error("Request must be set")
                
                socket_data = "action=%s\n\n" % self.smtpd_access_policy(attrs)
                connection.sendall(socket_data)
        except socket.timeout:
            try:
                connection.sendall("Timeout. Bye\n")
            except:
                self.logger.debug('Fail to send bye')
        except:
            self.logger.exception("Problem while handling request")
        finally:
            self.logger.debug("Closing socket")
            connection.close()


# Handle user-defined signals
# -----------------------------------------------------------------------------

    def handle_signal(self, signum, frame):
        if signum == signal.SIGUSR1:
            self.logger.debug("Signal %s triggered - flushing cache" % signum)
            self.datamanager.flush()
        elif signum == signal.SIGUSR2:
            self.logger.debug("Signal %s triggered - reloading configuration" % signum)
            self.load_config()
        elif signum == signal.SIGHUP:
            self.logger.debug("Signal %s triggered - reloading cache" % signum)
            self.datamanager.load()
        elif signum == signal.SIGINT or signum == signal.SIGTERM:
            self.logger.debug("Signal %s triggered - flush then stop daemon" % signum)
            self.datamanager.flush()
            self.stop()
        else:
            self.logger.debug("Unknown signal %s triggered - pass" % signum)


# Check service running state 
# -----------------------------------------------------------------------------

    def check_connections(self):
        netstat_output = os.popen("netstat -lntp").read()

        policy_server_active = True if netstat_output.count("9001") > 0 else False
        api_server_active = True if netstat_output.count("9002") > 0 else False

        return {
            'policy': policy_server_active,
            'api': api_server_active
        }


# Entrance of the request, apply regex on input stream
# -----------------------------------------------------------------------------

    def parse_request(self, connection, data):
        attrs = {}
        dataset = data.split("\n")

        for val in dataset:
            # necessary to match data values with method names of DataManager
            matched = re.compile("([^=]+)=(.*)").match(val)

            if matched:
                # get matching pattern groups and apply a substring on them
                (attr_key, attr_val) = matched.group(1, 2)
                attrs[attr_key.strip("\r\n\\ ")] = attr_val.strip("\r\n")

                self.logger.debug("Attribute key: %s - attribute value: %s" % (attr_key, attr_val))
            elif data == "\n":
                self.logger.debug("Arguments hash filled -> executing policy")
                break
            else:
                self.logger.debug("Unknown input, we pass")

        return attrs


# Handle treatments relative to policy access
# -----------------------------------------------------------------------------

    def smtpd_access_policy(self, attrs):
        self.logger.debug("Handle mail policy")
        
        action = "DUNNO"
        messages = []
        increment = True

        user_with_fqdn = attrs.get("sasl_username")
        
        # we need to clean username here, to remove the domain part
        matched = re.compile("([^@]+)").match(user_with_fqdn)
        user = matched.group(1)

        if user is None:
            self.logger.info("No user, so the mail shall pass" % user)
            return ""

        for period in self.periods:
            count = self.datamanager.retrieve(user, period, "count")
            limit = self.ldapmanager.get_limit(user, period)

            if int(limit) == -2:
                return "REJECT Service temporarily unavailable"
            # display "unlimited" instead of "-1"
            elif int(limit) == -1:
                limit_str = "unlimited"
            else:
                limit_str = limit

            # limit reached
            if ((int(limit) > -1) and (int(count) >= int(limit))):
                increment = False
                action = "REJECT forbidden"
                messages.append("[%s: NO (%s/%s)]" % (period, count, limit_str))
                self.logger.debug("Data for period: %s -- count:%s & limit:%s" % (period, count, limit_str))
                break
            else:
                messages.append("[%s: YES (%s/%s)]" % (period, count, limit_str))
                self.logger.debug("Data for period: %s -- count:%s & limit:%s" % (period, count, limit_str))

        # if an increment has been done previously, we need to try to save cache
        if increment:
            self.datamanager.increment(user)
            self.logger.info("Mail has been accepted -- %s" % " - ".join(messages))
        else:
            self.logger.info("Mail has been rejected -- %s" % " - ".join(messages))

        self.datamanager.save()

        return action + " -- " + " - ".join(messages)

# Start server
# -----------------------------------------------------------------------------

    def start(self):
        # check 
        states = server.check_connections()

        if states.get("policy"):
            self.logger.critical("Policy socket already bound, giving up...")
            sys.exit(3)
        elif states.get("api"): 
            self.logger.critical("Api socket already bound, giving up...")
            sys.exit(4)

        self.logger.debug("Starting daemon, please wait...")

        # load the managers
        self.datamanager = DataManager(self)
        self.ldapmanager = LdapManager(self)

        # load api manager in a separate thread
        self.api_manager = ApiManager(self)
        api_thread = Thread(target=self.api_manager.start, args=())
        api_thread.start()
        self.threads.append(api_thread)

        # run main server
        try:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            self.socket.bind((self.hostname, self.port))

            signal.signal(signal.SIGTERM, self.handle_signal)
            signal.signal(signal.SIGINT, self.handle_signal)
            signal.signal(signal.SIGUSR1, self.handle_signal)
            signal.signal(signal.SIGUSR2, self.handle_signal)
            signal.signal(signal.SIGHUP, self.handle_signal)

            self.socket.listen(1)

            while True:
                try:
                    conn, address = self.socket.accept()
                    self.logger.debug("Got new connection")
                    t = Thread(target=self.handle_connection, args=(conn, address))
                    t.start()
                    self.threads.append(t)
                    self.logger.debug("Started thread %r", t)
                except socket.error, e:

                    # Catch Interrupted system call
                    if e[0] == 4:
                        continue
                    # Do nothing if server is already stop
                    if not self.stopped:
                        self.logger.debug("Unexcepted exception : %s" % e)
                        self.stop()
                    break
        except socket.error, e:
            self.logger.critical("Problem with socket connection: %s" % e)
            self.stop()


# Stop server
# -----------------------------------------------------------------------------

    def stop(self):
        # check 
        states = server.check_connections()

        if states.get("api"): 
            self.api_manager.stop()
            self.logger.debug("The API server is stopped, please wait...")

        if states.get("policy"):
            self.logger.debug("Stopping daemon, please wait...")

            self.socket.close()
            self.logger.debug("Socket is closed, please wait...")
            
            for thread in self.threads:
                thread.join()
            self.logger.debug("All threads are closed, please wait...")

        pidfile = self.config.get("global", {}).get("pid_file")

        if pidfile is not None:
            if os.path.isfile(pidfile):
                os.unlink(pidfile)
                self.logger.debug("Delete pid file...")

        self.stopped = True
        self.logger.debug("The daemon has been fully stopped")


# Main method
# -----------------------------------------------------------------------------

# FIXME sortie if not root

if __name__ == "__main__":
    server = Server()

    # handle cli arguments
    parser = argparse.ArgumentParser(prog='policy_manager')

    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('-d', '--daemonize', action='store_true')
    parser.add_argument('-c', '--configuration-file')

    args = parser.parse_args()
    server.args = args.__dict__

    # load and propagate configuration
    server.load_config()

    # run server
    server.start()
